#!/usr/bin/env python
"""
Cuckoo search.

Usage:
  cs.py [options]

Options:
  -h, --help  Show this message and exit.
  -n N        Number of generations in CS. [default: 20]
  --print-level LEVEL
              Print verbose level. [default: 1]
  --update-vrange NUPDATE
              Update variable range per N. [default: -1]
"""
import os
import sys
from docopt import docopt
import numpy as np
from numpy import exp, sin, cos
import random
import copy
from multiprocessing import Process, Pool
from time import time
from scipy.special import gamma
import pandas as pd

from optzer.individual import Individual, ind_from_db, update_slims
from optzer.testfunc import testfunc, write_vars_for_testfunc
from optzer.io import write_db_optzer, read_db_optzer

__author__ = "RYO KOBAYASHI"
__version__ = "221227"

_fname_gen = 'out.cs.generations'
_fname_ind = 'out.cs.individuals'
_fname_db = 'db.optzer.json'


class CS:
    """
    Cuckoo search class.
    """

    def __init__(self, nind, frac, vnames, vs0, slims,
                 hlims, loss_func, write_func=None,
                 nproc=0, seed=42, **kwargs):
        """
        Conctructor of CS class.

        nind:  Number of individuals.
        frac:  Fraction of worse individuals to be abondoned.
        loss_func:  function
            Loss function to be minimized with variables and **kwargs.
        nproc:  int
            Number of processes used to run N individuals.
        vnames:  list
            Variable names.
        vs0: dict
            Initial guess of variables.
        slims,hlims: dict
            Set of variables with names.
        """
        if nind < 2:
            raise ValueError('nind must be greater than 1 in CS!')
        np.random.seed(seed)
        random.seed(seed)
        self.nind = nind   # Number of individuals in a generation
        self.frac = frac   # Fraction of worse individuals to be abondoned
        self.nproc = nproc
        self.vnames = vnames
        self.vs0 = vs0
        self.slims = slims
        self.hlims = hlims
        self.vws = {}
        for k in vnames:
            self.vws[k] = max(self.slims[k][1] -self.slims[k][0], 0.0)
        self.loss_func = loss_func
        self.write_func = write_func
        self.kwargs = kwargs
        self.bestind = None
        self.print_level = 0
        if 'print_level' in kwargs.keys():
            self.print_level = int(kwargs['print_level'])
        
        if 'update_vrange' in kwargs.keys():
            self.update_slims_per = kwargs['update_vrange']
        else:
            self.update_slims_per = -1

        self.beta = 1.5
        self.betai = 1.0 /self.beta
        self.usgm = (gamma(1+self.beta)*np.sin(np.pi*self.beta/2)/ \
                     gamma((1+self.beta)/2)*self.beta*2.0**((self.beta-1)/2))**self.betai
        self.vsgm = 1.0

        cols = ['iid','loss','gen'] +self.vnames
        self.history_db = pd.DataFrame(columns=cols)
        #...Try to restart by loading db.optzer.json if exists
        if os.path.exists(_fname_db):
            try:
                self.history_db = pd.read_json(_fname_db)
            except Exception as e:
                print(e)
                print(f'\n !!! Failed to load {_fname_db} for restart. !!!'
                      +'\n !!! So start with the given initial guess.     !!!')
            print(f'\n Restarting with existing DB, {_fname_db}.')
                
        #...initialize population
        self.population = []
        if len(self.history_db) == 0:
            self.igen0 = 0
            self.iidinc = 0
            for i in range(self.nind):
                self.iidinc += 1
                ind = Individual(self.iidinc, self.vnames)
                if i == 0:
                    ind.set_variables(self.vs0, self.slims)
                else:
                    ind.init_random(self.slims)
                self.population.append(ind)
        else:  # DB loaded
            #...Use DB to narrow the soft limits
            self.slims = update_slims(self.slims,self.hlims,
                                      self.history_db)
            print(' Update variable ranges')
            for k in self.vnames:
                print(' {0:>15s}:  {1:7.3f}  {2:7.3f}'.format(k,
                                                              self.slims[k][0],
                                                              self.slims[k][1]))
            #...Get best one from the DB
            bestidx = self.history_db.loss.argmin()
            self.bestind = ind_from_db(self.history_db,
                                       bestidx,
                                       self.vnames, self.slims)
            #...Generate random population with the updated slims
            self.iidinc = int(self.history_db.iid.max()) +1
            self.igen0 = int(self.history_db.gen.max()) +1
            for i in range(self.nind):
                ind = Individual(self.iidinc, self.vnames)
                ind.init_random(self.slims)
                self.population.append(ind)
                self.iidinc += 1
                

        #...Evaluate loss function values
        prcs = []
        if self.nproc > 0 :  # use specified number of cores by nproc
            pool = Pool(processes=self.nproc)
        else:
            pool = Pool()
            
        for ip,ind in enumerate(self.population):
            kwtmp = copy.copy(self.kwargs)
            kwtmp['index'] = ip
            kwtmp['iid'] = ind.iid
            prcs.append(pool.apply_async(ind.calc, (self.loss_func,kwtmp,)))
        results = [ res.get() for res in prcs ]
        for res in results:
            loss,ip = res
            self.population[ip].loss = loss
            self.population[ip].gen = self.igen0

        pool.close()
        
        self.keep_best()
        #...Create history DB if not exists
        pop2dfs = [ pi.to_DataFrame() for pi in self.population ]
        self.history_db = pd.concat([self.history_db] +pop2dfs,
                                    ignore_index=True)
            
        # if self.print_level > 2:
        #     for ind in self.population:
        #         fname = 'in.vars.optzer.{0:d}'.format(ind.iid)
        #         self.write_variables(ind,
        #                              fname=fname,
        #                              **self.kwargs)
        # else:
        #     fname = 'in.vars.optzer.{0:d}'.format(self.bestind.iid)
        #     self.write_variables(self.bestind,
        #                          fname=fname,
        #                          **self.kwargs)

        #...Once the history is updated, dump it to the file
        write_db_optzer(self.history_db,fname=_fname_db)
        return None

    def keep_best(self):
        losses = []
        for i,pi in enumerate(self.population):
            if pi.loss == None:
                raise ValueError('Something went wrong.')
            losses.append(pi.loss)

        minloss = min(losses)
        if self.bestind == None or minloss < self.bestind.loss:
            idx = losses.index(minloss)
            self.bestind = copy.deepcopy(self.population[idx])
        return None

    def sort_individuals(self):
        """
        Sort individuals in the population in the ascending order.
        """
        jtop = self.nind
        for i in range(self.nind):
            jtop -= 1
            for j in range(jtop):
                pj = self.population[j]
                pjp = self.population[j+1]
                if pj.loss > pjp.loss:
                    self.population[j] = pjp
                    self.population[j+1] = pj
        return None
        
    def run(self,max_gen=100):
        """
        Perfom CS.
        """

        if 'start' in self.kwargs.keys():
            start = self.kwargs['start']
        else:
            start = time()

        if self.print_level > 0:
            print(' step,time,best_iid,best_loss,vars='
                  +' {0:6d} {1:8.1f} {2:5d} {3:8.4f}'.format(self.igen0,
                                                             time()-start,
                                                             self.bestind.iid,
                                                             self.bestind.loss),
                  end="")
            inc = 0
            for k in self.vnames:
                if inc < 16:
                    print(' {0:6.3f}'.format(self.bestind.vs[k]),end="")
                else:
                    break
                inc += 1
            print('', flush=True)

        #...Create pool before going into max_gen-loop,
        #...since creating pool inside could cause "Too many files" error.
        if self.nproc > 0 :  # use specified number of cores by nproc
            pool = Pool(processes=self.nproc)
        else:
            pool = Pool()
            
        for igen in range(self.igen0, self.igen0+max_gen):
            self.sort_individuals()
            #...Create candidates from current population using Levy flight
            candidates = []
            vbest = self.bestind.vs
            for ip,pi in enumerate(self.population):
                vi = pi.vs
                vnew = copy.copy(vi)
                for k in self.vnames:
                    u = np.random.normal()*self.usgm
                    v = abs(np.random.normal()*self.vsgm)
                    v = max(v,1.0e-8)
                    w = u/v**self.betai
                    zeta = self.vws[k] *0.01 *w
                    # zeta = self.vws[iv]*0.01 *w *(vi[iv] -vbest[iv])
                    # if ip == 0:
                    #     zeta = self.vws[iv] *0.001 *w
                    # else:
                    #     zeta = 0.01 *w *(vi[iv] -vbest[iv])
                    vnew[k] = vnew[k] +zeta*np.random.normal()
                #...create new individual for trial
                # print('ip,vi,vnew=',ip,vi,vnew)
                self.iidinc += 1
                newind = Individual(self.iidinc, self.vnames)
                newind.set_variables(vnew, self.slims)
                candidates.append(newind)

            #...Create new completely random candidates
            iab = int((1.0 -self.frac)*self.nind)
            rnd_candidates = []
            for iv in range(iab,self.nind):
                self.iidinc += 1
                newind = Individual(self.iidinc, self.vnames)
                newind.init_random(self.slims)
                rnd_candidates.append(newind)

            #...Evaluate loss function values of updated candidates and new random ones
            prcs = []
            for ic,ci in enumerate(candidates):
                kwtmp = copy.copy(self.kwargs)
                kwtmp['index'] = ic
                kwtmp['iid'] = ci.iid
                prcs.append(pool.apply_async(ci.calc, (self.loss_func,kwtmp,)))
            rnd_prcs = []
            for ic,ci in enumerate(rnd_candidates):
                kwtmp = copy.copy(self.kwargs)
                kwtmp['index'] = len(candidates) +ic
                kwtmp['iid'] = ci.iid
                rnd_prcs.append(pool.apply_async(ci.calc, (self.loss_func,kwtmp,)))
            
            results = [ res.get() for res in prcs ]
            rnd_results = [ res.get() for res in rnd_prcs ]

            for res in results:
                loss,ic = res
                candidates[ic].loss = loss
                candidates[ic].gen = igen+1
            db_add = [ c.to_DataFrame() for c in candidates ] 

            for res in rnd_results:
                loss,ic_rnd = res
                ic = ic_rnd -len(candidates)
                rnd_candidates[ic].loss = loss
                rnd_candidates[ic].gen = igen+1
            db_add.extend([ c.to_DataFrame() for c in rnd_candidates ])
            self.history_db = pd.concat([self.history_db] +db_add)
            self.history_db.drop_duplicates(subset='iid',
                                            inplace=True,
                                            ignore_index=True)
            #...Dump the DB once it is updated.
            write_db_optzer(self.history_db,fname=_fname_db)

            #...Pick j that is to be compared with i
            js = random.sample(range(self.nind),k=self.nind)
            #...Decide whether or not to adopt new one
            for jc,jv in enumerate(js):
                pj = self.population[jv]
                cj = candidates[jc]
                dloss = cj.loss -pj.loss
                if dloss < 0.0:  # replace with new individual
                    self.population[jv] = cj
                else:
                    pass

            #...Rank individuals
            self.sort_individuals()
            
            #...Replace to-be-abandoned ones with new random ones
            ic = 0
            for iv in range(iab,self.nind):
                ci = rnd_candidates[ic]
                ic += 1
                self.population[iv] = ci

            #...Check best
            best_updated = False
            for ic,ci in enumerate(self.population):
                if ci.loss < self.bestind.loss:
                    self.bestind = ci
                    best_updated = True
            if best_updated:
                self.write_variables(self.bestind,
                                     fname='in.vars.optzer.best',
                                     **self.kwargs)

            #...Update variable ranges if needed
            if self.update_slims_per > 0 and (igen+1) % self.update_slims_per == 0:
                self.slims = update_slims(self.slims,self.hlims,
                                          self.history_db)
                print(' Update variable ranges')
                for k in self.vnames:
                    print(' {0:>10s}:  {1:7.3f}  {2:7.3f}'.format(k,self.slims[k][0],self.slims[k][1]))
            
            if self.print_level > 0:
                print(' step,time,best_iid,best_loss,vars='
                      +' {0:6d} {1:8.1f} {2:5d} {3:8.4f}'.format(igen+1,time()-start,
                                                                 self.bestind.iid,
                                                                 self.bestind.loss),end="")

                inc = 0
                for k in self.vnames:
                    if inc < 16:
                        print(' {0:6.3f}'.format(self.bestind.vs[k]),end="")
                    else:
                        break
                    inc += 1
                print('', flush=True)

        pool.close()
        #...Finaly write out the best one
        self.write_variables(self.bestind,
                             fname='in.vars.optzer.best',
                             **self.kwargs)
        return None

    def write_variables(self,ind,fname='in.vars.optzer',**kwargs):
        if self.write_func != None:
            self.write_func(ind.vnames, ind.vs, self.slims, self.hlims,
                            fname, **kwargs)
        return None

def main():
    args = docopt(__doc__.format(os.path.basename(sys.argv[0])))
    ngen = int(args['-n'])
    kwargs = {}
    kwargs['print_level'] = int(args['--print-level'])
    kwargs['update_vrange'] = int(args['--update-vrange'])

    vnames = ['x','y']
    vs = { 'x':1.0, 'y':-0.5 }
    slims = { 'x':[-1.0, 2.0],'y':[-1.0, 1.0] }
    hlims = { 'x':[-2.0, 2.0],'y':[-2.0, 2.0] }
    
    cs = CS(10, 0.25, vnames, vs, slims, hlims,
            testfunc, write_vars_for_testfunc, **kwargs)
    cs.run(ngen)
    return None
    

if __name__ == "__main__":

    main()



#!/usr/bin/env python
"""
Tree-based Parzen Estimator (TPE).

Usage:
  {0:s} [options]

Options:
  -h, --help  Show this message and exit.
"""
import os, sys
from docopt import docopt
import numpy as np
import random
import copy
from multiprocessing import Process, Pool
from time import time
import pandas as pd

from optzer.individual import Individual, ind_from_db
from optzer.io import write_db_optzer, read_db_optzer

__author__ = "RYO KOBAYASHI"
__version__ = "221228"

_fname_db = 'db.optzer.json'

def gauss_kernel(x):
    return np.exp(-0.5*x*x)/np.sqrt(2.0*np.pi)

class Sample:
    """
    A sample point that has parameter vector and loss value.
    """
    def __init__(self, iid, ndim, loss_func):
        self.iid = iid
        self.ndim = ndim
        self.loss_func = loss_func
        self.variables = np.zeros(self.ndim)
        self.val = None
        return None

    def set_variables(self,variables):
        if len(variables) != len(self.variables):
            raise ValueError('len(variables) != len(self.variables)')

        self.variables[:] = variables[:]
        self.val = None
        return None

    def calc_loss_func(self, vlogs, **kwargs):
        """
        Compute loss function value using self.loss_func function given in the constructor.
        """
        vec = copy.copy(self.variables)
        # If vlog != None, some of variables may be expressed in log domain,
        # they must be transformed back to non-log domain.
        for i in range(len(vec)):
            if vlogs[i]:
                vec[i] = np.exp(vec[i])
        val = self.loss_func(vec, **kwargs)
        return val,kwargs['index']

class TPE:
    """
    Class for Tree-based Parzen Estimator (TPE) or Weighted Parzen Estimator (WPE).
    
    Refs:
      1. Bergstra, J., Bardenet, R., Bengio, Y. & Kégl, B.  in Proc. NIPS-24th, 2546–2554
    """

    def __init__(self, nbatch, vnames, vs0, slims, hlims, loss_func,
                 write_func=None, seed=42, **kwargs):
        """
        Conctructor of TPE class.

        Parameters:
          nbatch : int
              Number of samples in a batch which is the same as num of processes.
          vs0 : dict
              Initial guess of variables.
          slims : dict
              Current lower and upper limit of variables, which are used only at random sampling processes.
          hlims : dict
              Hard limits of variables, which are fixed during TPE iterations. 
          loss_func : function
              Loss function to be minimized with variables and **kwargs.
          write_func : function
              Function for outputing some info.
        """
        if nbatch < 1:
            raise ValueError('nbatch must be > 0.')
        np.random.seed(seed)
        random.seed(seed)
        self.nbatch = nbatch
        self.ndim = len(vs0)
        self.vnames = vnames
        self.vs0 = vs0
        self.hlims = hlims
        #self.slims = slims
        self.slims = hlims  # Since slims is not used, use hlims instead as slims
        self.vlogs = { k:False for k in vnames }
        if 'vlogs' in kwargs.keys():
            self.vlogs = kwargs['vlogs']
        self.loss_func = loss_func
        self.write_func = write_func
        self.kwargs = kwargs
        self.best_pnt = None
        self.print_level = 0
        self.nsmpl_prior = 100
        self.ntrial = 100
        self.method = kwargs['opt_method']

        self.gamma = 0.15
        #...Change default values if specified
        if 'print_level' in kwargs.keys():
            self.print_level = int(kwargs['print_level'])
        if 'tpe_nsmpl_prior' in kwargs.keys():
            self.nsmpl_prior = int(kwargs['tpe_nsmpl_prior'])
        if 'tpe_ntrial' in kwargs.keys():
            self.ntrial = int(kwargs['tpe_ntrial'])
        if 'tpe_gamma' in kwargs.keys():
            self.gamma = float(kwargs['tpe_gamma'])

        if self.gamma < 0.0 or self.gamma > 1.0:
            raise ValueError('gamma must be within 0. and 1., whereas gamma = ',self.gamma)

        #...Write info
        print('')
        if self.method in ('wpe','WPE'):
            print(' {0:s} infomation:'.format(self.method))
            print(f'   Num of prior samples = {self.nsmpl_prior:d}')
            print(f'   Num of top samples used for density estimation = {self.ntrial:d}')
        elif self.method in ('tpe','TPE'):
            print(' {0:s} infomation:'.format(self.method))
            print(f'   Num of prior samples = {self.nsmpl_prior:d}')
            print(f'   Num of trials for sampling = {self.ntrial:d}')
            print(f'   Gamma for dividing high and low = {self.gamma:4.2f}')
        print('')
    
        #...Change vrange if log domain
        for k,v in self.vlogs.items():
            if not v: continue
            self.vs0[k] = np.log(self.vs0[k])
            for l in range(2):
                self.slims[k][l] = np.log(self.slims[k][l])
                self.hlims[k][l] = np.log(self.hlims[k][l])
        # for i in range(self.ndim):
        #     if self.vlogs[i]:
        #         self.vs0[i] = np.log(self.vs0[i])
        #         for l in range(2):
        #             self.slims[i,l] = np.log(self.slims[i,l])
        #             self.hlims[i,l] = np.log(self.hlims[i,l])

        cols = ['iid','loss','gen'] +self.vnames
        self.history_db = pd.DataFrame(columns=cols)  # History of all samples
        #...Try to restart by loading db.optzer.json if exists
        if os.path.exists(_fname_db):
            try:
                self.history_db = pd.read_json(_fname_db)
            except Exception as e:
                print(e)
                print(f'\n !!! Failed to load {_fname_db} for restart. !!!'
                      +'\n !!! So start with the given initial guess.     !!!')
            print(f'\n Restarting with existing DB, {_fname_db}.')

        #...Initialize sample history
        self.candidates = []
        self.igen0 = 0
        self.bestsmpl = None
        if len(self.history_db) == 0:
            self.iidinc = 0
            for i in range(self.nbatch):
                self.iidinc += 1
                smpl = Individual(self.iidinc, self.vnames )
                if i == 0:
                    smpl.set_variables(self.vs0, self.slims)
                else:
                    smpl.init_random(self.slims)
                self.candidates.append(smpl)
        else:  # DB loaded
            #...Get best one from the DB
            bestidx = self.history_db.loss.argmin()
            self.bestsmpl = ind_from_db(self.history_db,
                                        bestidx,
                                        self.vnames, self.slims)
            #...Generate random population with the updated slims
            self.iidinc = int(self.history_db.iid.max()) +1
            self.igen0 = int(self.history_db.gen.max()) +1
            for i in range(self.nbatch):
                ind = Individual(self.iidinc, self.vnames)
                ind.init_random(self.slims)
                self.candidates.append(ind)
                self.iidinc += 1
            
        return None

    def keep_best(self):
        losses = []
        for i,si in enumerate(self.candidates):
            if si.loss == None:
                raise ValueError('Something went wrong.')
            losses.append(si.loss)

        minloss = min(losses)
        if self.bestsmpl == None or minloss < self.bestsmpl.loss:
            idx = losses.index(minloss)
            self.bestsmpl = copy.deepcopy(self.candidates[idx])
        return None

    def run(self,maxstp=0):
        """
        Perfom TPE.
        """

        starttime = time()

        #...Create pool before going into maxstp-loop,
        #...since creating pool inside could cause "Too many files" error.
        pool = Pool(processes=self.nbatch)

        #...Evaluate sample losses of initial sets
        prcs = []
        for i,ci in enumerate(self.candidates):
            kwtmp = copy.copy(self.kwargs)
            kwtmp['index'] = i
            kwtmp['iid'] = ci.iid
            prcs.append(pool.apply_async(ci.calc, (self.loss_func,kwtmp,)))

        results = [ res.get() for res in prcs ]
        for res in results:
            loss, i = res
            self.candidates[i].loss = loss
            self.candidates[i].gen = self.igen0
        db_add = [ ci.to_DataFrame() for ci in self.candidates ]
        self.history_db = pd.concat([self.history_db] +db_add)
        self.history_db.drop_duplicates(subset='iid',
                                        inplace=True,
                                        ignore_index=True)
        #...Dump the DB once it is updated.
        write_db_optzer(self.history_db,fname=_fname_db)

        # self.slims = update_slims(self.slims, self.hlims, self.history_db)
        
        #...Check best
        self.keep_best()
        self.write_variables(self.bestsmpl,
                             fname='in.vars.optzer.best',
                             **self.kwargs)
        
        if self.print_level > 0:
            self._write_step_info(self.igen0,starttime)

        #...TPE loop starts
        for igen in range(self.igen0+1, self.igen0+1+maxstp):
            #...Create candidates by either random or TPE
            if len(self.history_db) <= self.nsmpl_prior:
                #...Create random candidates
                self.candidates = []
                for i in range(self.nbatch):
                    self.iidinc += 1
                    newsmpl = Individual(self.iidinc, self.vnames)
                    newsmpl.init_random(self.slims)
                    self.candidates.append(newsmpl)
            else:
                if self.method in ('wpe,''WPE'):
                    #...Create candidates by WPE
                    self.candidates = self._candidates_by_WPE()
                else:
                    #...Create candidates by TPE
                    self.candidates = self._candidates_by_TPE()

            #...Evaluate sample losses
            prcs = []
            for i,ci in enumerate(self.candidates):
                kwtmp = copy.copy(self.kwargs)
                kwtmp['index'] = i
                kwtmp['iid'] = ci.iid
                prcs.append(pool.apply_async(ci.calc, (self.loss_func,kwtmp,)))

            results = [ res.get() for res in prcs ]
            for res in results:
                loss, i = res
                self.candidates[i].loss = loss
                self.candidates[i].gen = igen+1
            db_add = [ c.to_DataFrame() for c in self.candidates ]
            self.history_db = pd.concat([self.history_db] +db_add)
            self.history_db.drop_duplicates(subset='iid',
                                            inplace=True,
                                            ignore_index=True)
            #...Dump the DB once it is updated.
            write_db_optzer(self.history_db,fname=_fname_db)

            # self.slims = update_slims(self.slims, self.hlims, self.history_db)

            #...Check best
            best_updated = False
            for ip,ci in enumerate(self.candidates):
                if ci.loss < self.bestsmpl.loss:
                    self.bestsmpl = ci
                    best_updated = True
            if best_updated:
                self.write_variables(self.bestsmpl,
                                     fname='in.vars.optzer.best',
                                     **self.kwargs)

            #...Write info
            if self.print_level > 0:
                self._write_step_info(igen,starttime)
            
        pool.close()
        return None

    def _write_step_info(self,istp,starttime):
        print(' step,time,best_iid,best_loss,vars='
              +' {0:6d} {1:8.1f} {2:5d} {3:8.4f}'.format(istp+1,
                                                         time()-starttime,
                                                         self.bestsmpl.iid,
                                                         self.bestsmpl.loss),end="")
        inc = 0
        for k in self.vnames:
            if inc < 16:
                print(' {0:6.3f}'.format(self.bestsmpl.vs[k]),end="")
            else:
                break
            inc += 1
        print('', flush=True)
        return None

    def _candidates_by_TPE(self,):
        """
        Create candidates by using TPE.
        """
        losses = self.history_db.loss
        nlow = int(self.gamma *len(losses))
        nhigh = len(losses) -nlow
        argpart = np.argpartition(losses,nlow)
        Xlow = np.zeros((nlow,self.ndim))
        Xhigh = np.zeros((nhigh,self.ndim))
        for i in range(nlow):
            idx = argpart[i]
            vs = ind_from_db(self.history_db,idx,self.vnames,self.slims).vs
            for j,k in enumerate(self.vnames):
                Xlow[i,j] = vs[k]
        for i in range(nhigh):
            idx = argpart[nlow+i]
            vs = ind_from_db(self.history_db,idx,self.vnames,self.slims).vs
            for j,k in enumerate(self.vnames):
                Xhigh[i,j] = vs[k]

        #...Sampling variable candidates
        xcandidates = np.empty((self.nbatch,self.ndim))
        ntrial = max(self.ntrial,self.nbatch)
        for idim in range(self.ndim):
            key = self.vnames[idim]
            xhmin = self.hlims[key][0]
            xhmax = self.hlims[key][1]
            xlowsrt = np.sort(Xlow[:,idim])
            npnt = len(xlowsrt)
            #...Determine smoothness parameter, h, by Silverman's method
            q75, q25 = np.percentile(xlowsrt, [75,25])
            std = np.std(xlowsrt)
            sgm = min(std, (q75-q25)/1.34)
            h = 1.06 *sgm /np.power(npnt, 1.0/5)
            #...Prepare for g(x)
            xhighsrt = Xhigh[:,idim]
            q75, q25 = np.percentile(xhighsrt, [75,25])
            sgmh = min(np.std(xhighsrt), (q75-q25)/1.34)
            hh = 1.06 *sgmh /np.power(len(xhighsrt),1.0/5)
            #...Several trials for selection
            aquisition = np.zeros(ntrial)
            xs = np.empty(ntrial)
            for itry in range(ntrial):
                ipnt = int(np.random.random()*npnt)
                xi = xlowsrt[ipnt]
                r = h *np.sqrt(-2.0*np.log(np.random.random()))
                th = 2.0 *np.pi *np.random.random()
                x = xi + r*np.cos(th)
                #...Wrap by vranges
                x = min(max(x,xhmin),xhmax)
                xs[itry] = x
                #...Compute l(x) and g(x)
                lx = 0.0
                for j in range(npnt):
                    z = (x-xlowsrt[j])/h
                    lx += np.exp(-0.5*z*z)
                lx /= npnt*h *np.sqrt(2.0*np.pi)
                gx = 0.0
                for j in range(len(xhighsrt)):
                    z = (x-xhighsrt[j])/hh
                    gx += np.exp(-0.5*z*z)
                gx /= len(xhighsrt)*hh *np.sqrt(2.0*np.pi)
                aquisition[itry] = gx/lx
            #...Pick nbatch of minimum aquisition points
            idxsort = np.argsort(aquisition)
            xcandidates[:,idim] = xs[idxsort[0:self.nbatch]]
        #...Create sample with xcandidate as variables
        candidates = []
        for ib in range(self.nbatch):
            self.iidinc += 1
            smpl = Individual(self.iidinc, self.vnames)
            vsnew = {}
            for j,k in enumerate(self.vnames):
                vsnew[k] = xcandidates[ib,j]
            smpl.set_variables(vsnew, self.slims)
            candidates.append(smpl)
        return candidates

    def _candidates_by_WPE(self,):
        """
        Create candidates by using WPE.
        """
        losses = self.history_db.loss
        if len(self.history_db) > self.ntrial:
            iargs = np.argpartition(losses, self.ntrial)
            #tmpsmpls = [ self.history[i] for i in iargs[:self.ntrial] ]
            tmpsmpls = [ ind_from_db(self.history_db,idx,self.vnames,self.slims)
                         for idx in iargs[:self.ntrial] ]
            losses = losses[ iargs[:self.ntrial] ]
        else:
            #tmpsmpls = copy.copy(self.history)
            tmpsmpls = [ ind_from_db(self.history_db,idx,self.vnames,self.slims)
                         for idx in self.history_db.index ]
        vmin = losses.min()
        wgts = np.array([ np.exp(-(v-vmin)/vmin) for v in losses ])
        xtmps = np.zeros((len(tmpsmpls),self.ndim))
        for i in range(len(tmpsmpls)):
            for j,k in enumerate(self.vnames):
                xtmps[i,j] = tmpsmpls[i].vs[k]
        #...Sampling variable candidates
        xcandidates = np.empty((self.nbatch,self.ndim))
        for idim in range(self.ndim):
            key = self.vnames[idim]
            xhmin = self.hlims[key][0]
            xhmax = self.hlims[key][1]
            xsrt = np.sort(xtmps[:,idim])
            #...Determine smoothness parameter, h, by Silverman's method
            q75, q25 = np.percentile(xsrt, [75,25])
            sgm = min(np.std(xsrt), (q75-q25)/1.34)
            h = 1.06 *sgm /np.power(len(xsrt), 1.0/5)
            xs = np.empty(self.nbatch)
            for ib in range(self.nbatch):
                ipnt = random.choices([j for j in range(len(wgts))],weights=wgts)[0]
                xi = xtmps[ipnt,idim]
                r = h *np.sqrt(-2.0*np.log(np.random.random()))
                th = 2.0 *np.pi *np.random.random()
                x = xi + r*np.cos(th)
                #...Wrap by slims
                x = min(max(x,xhmin),xhmax)
                xs[ib] = x
            xcandidates[:,idim] = xs[:]
        
        #...Create sample with xcandidate as variables
        candidates = []
        for ib in range(self.nbatch):
            self.iidinc += 1
            smpl = Individual(self.iidinc, self.vnames)
            vsnew = {}
            for j,k in enumerate(self.vnames):
                vsnew[k] = xcandidates[ib,j]
            smpl.set_variables(vsnew, self.slims)
            candidates.append(smpl)
        
        return candidates

    def write_variables(self,smpl,fname='in.vars.optzer',**kwargs):
        if self.write_func != None:
            self.write_func(smpl.vnames, smpl.vs, self.slims, self.hlims,
                            fname, **kwargs)
        return None

def main():
    args = docopt(__doc__.format(os.path.basename(sys.argv[0])))
    return None

if __name__ == "__main__":

    main()



#!/usr/bin/env python
"""
Get best data from the directory that optzer was performed.

Usage:
  {0:s} [options]

Options:
  -h, --help  Show this message and exit.
  -o,--out-file OUTFILE
              Specify output file name. [default: out.optzer]
  --bestdir-name BESTDIR
              Best data dir name. [default: best_data]
"""
import os,sys
import shutil
from docopt import docopt

from .io import read_out_optzer

__author__ = "RYO KOBAYASHI"
__version__ = "122224"

def read_output(fname='out.optzer'):
    """Read output from optzer and return some information."""
    with open(fname,'r') as f:
        lines = f.readlines()
    bestiid = -1
    for line in lines:
        if 'step,time' in line:
            data = line.split()
            bestiid = int(data[3])
            bestloss = float(data[4])
    return bestiid,bestloss

def main():
    args = docopt(__doc__.format(os.path.basename(sys.argv[0])))
    ofname = args['--out-file']
    bestdname = args['--bestdir-name']

    bestiid,_,_,_ = read_out_optzer(ofname)

    if os.path.exists(bestdname):
        shutil.rmtree(bestdname)

    # Look for the bestiid directory
    print(f' copying iid_{bestiid} ==> ./{bestdname}')
    cwd = os.getcwd()
    for path,dirs,files in os.walk(cwd):
        dname = os.path.basename(path)
        if dname == f'iid_{bestiid:d}':
            shutil.copytree(path, f'./{bestdname}')

    return None

if __name__ == "__main__":

    main()
